import fetch from 'dva/fetch';
import _ from 'lodash';
import { Resolver } from 'fastjson_ref_resolver';
import { getCookie, encrypt } from '../utils/helper';
import { cookie } from '../utils/config';
import { errors } from '../utils/error';

// eslint-disable-next-line max-len
function parseObject(response, { num2str = false, bool2str = false, nul2str = false, ud2str = false, nil2str = false } = {}) {
  if (response.status === 204) { // no-content
    return null;
  } else {
    const contentType = response.headers.get('content-type');
    if (contentType) {
      const needMap = num2str || bool2str || nul2str || ud2str || nil2str;
      const mapStr = (value) => {
        if (num2str && _.isNumber(value)) {
          return value.toString();
        }
        if (bool2str && _.isBoolean(value)) {
          return value.toString();
        }
        if (nul2str && _.isNull(value)) {
          return '';
        }
        if (ud2str && _.isUndefined(value)) {
          return '';
        }
        if (nil2str && _.isNil(value)) {
          return '';
        }
        return value;
      };
      const mapObj = (obj, mapArrFunc) => {
        if (_.isArray(obj)) {
          return mapArrFunc(obj, mapObj);
        }
        if (_.isPlainObject(obj)) {
          return _.mapValues(obj, (val) => {
            const ret = mapStr(val);
            return mapObj(ret, mapArrFunc);
          });
        }
        return obj;
      };
      const mapArr = (arr, mapObjFunc) => {
        if (_.isPlainObject(arr)) {
          return mapObjFunc(arr, mapArr);
        }
        if (_.isArray(arr)) {
          return _.map(arr, (val) => {
            const ret = mapStr(val);
            return mapArr(ret, mapObjFunc);
          });
        }
        return arr;
      };
      const mapValue = _.curry(mapObj)(_, mapArr);
      if (contentType.indexOf('json') !== -1) {
        const json = needMap ? response.json().then(mapValue) : response.json();
        return new Resolver(json).resolve();
      } else if (contentType.indexOf('xml') !== -1) {
        return response.text().then((text) => {
          return require.ensure([], (require) => {
            const { parseString } = require('xml2js');
            const options = {};
            const json = JSON.parse(parseString(text, options));
            return needMap ? mapValue(json) : json;
          });
        });
      } else if (contentType.indexOf('text') !== -1) {
        return response.text();
      } else {
        throw new Error(`Unsupported response content type: ${contentType}`);
      }
    } else {
      throw new Error('No response content type.');
    }
  }
}

function checkStatus(response) {
  if (response.status >= 200 && response.status < 300) {
    return response;
  }
  return response.json().then((data) => {
    const error = new Error(data.message ? data.message : response.statusText);
    error.data = data;
    throw error;
  });
}

const defaultOptions = {
  headers: { Accept: 'application/json' },
};

/**
 * Requests a URL, returning a promise.
 *
 * @param  {string} url       The URL we want to request
 * @param auth
 * @param params
 * @param  {object} [options] The options we want to pass to "fetch"
 * @return {object}           An object containing either "data" or "err"
 */
export default function request(url, params = {}, options = {}, auth = true) {
  let token;
  if (auth) {
    token = getCookie(cookie.token);
    if (!token) {
      return Promise.reject(errors.unLogin);
    }
    token = encrypt(token);
  }
  let queryParams = token ? [...normParams(params), ['token', token]] : normParams(params);
  queryParams = queryParams.map(([k, v]) => (_.isNil(v) ? k : `${k}=${encodeURIComponent(v)}`));
  queryParams = queryParams.join('&');
  let realUrl = url;
  if (queryParams) {
    realUrl = `${url}?${queryParams}`;
  }
  return fetch(realUrl, _.defaults(options, defaultOptions))
    .then(checkStatus)
    .then(parseObject);
}

export function normParams(unnormed) {
  if (_.isPlainObject(unnormed)) {
    return _(unnormed).toPairs().flatMap(([k, v]) => (_.isArray(v) ? v.map(vv => [k, vv]) : [[k, v]])).value();
  } else {
    return unnormed;
  }
}
